--[[
RealisticSeeder

Specialization for an realistc seeder

Author: 	Ifko[nator]
Date: 		04.07.2020
Version:	1.0

History:	v1.0 @04.07.2020 - initial implemation in FS 19
]]

RealisticSeeder = {};
RealisticSeeder.currentModName = g_currentModName;
RealisticSeeder.currentModDirectory = g_currentModDirectory;

function RealisticSeeder.prerequisitesPresent(specializations)
	return SpecializationUtil.hasSpecialization(FillUnit, specializations);
end;

function RealisticSeeder.registerEventListeners(vehicleType)
	local functionNames = {
		"onLoad",
		"onUpdateTick"
	};
	
	for _, functionName in ipairs(functionNames) do
		SpecializationUtil.registerEventListener(vehicleType, functionName, RealisticSeeder);
	end;
end;

function RealisticSeeder.registerFunctions(vehicleType)
	local newFunctions = {
		"loadSeedFillTypes",
		"getAttachedAdditionalTanks"
	};
	
	for _, newFunction in ipairs(newFunctions) do
		SpecializationUtil.registerFunction(vehicleType, newFunction, RealisticSeeder[newFunction]);
	end;
end;

function RealisticSeeder:onLoad(savegame)
	if RealisticSeederUtil == nil or not RealisticSeederUtil.getIsValidStoreCategory(getXMLString(self.xmlFile, "vehicle.storeData.category")) then
		return;
	end;
	
	local specRealisticSeeder = RealisticSeederUtil.getSpecByName(self, "realisticSeeder");
	local specSowingMachine = RealisticSeederUtil.getSpecByName(self, "sowingMachine");
	local specFillUnit = RealisticSeederUtil.getSpecByName(self, "fillUnit");

	local modDesc = loadXMLFile("modDesc", RealisticSeeder.currentModDirectory .. "modDesc.xml");

	RealisticSeeder.debugPriority = RealisticSeederUtil.getDebugPriority(modDesc, "modDesc.realisticSeeder#debugPriority");
	
	self:loadSeedFillTypes(modDesc, self.xmlFile, getXMLString(self.xmlFile, "vehicle.storeData.category"), getXMLString(self.xmlFile, "vehicle.base.size#width"));
	
	specRealisticSeeder.oldSeedFruitType = FillType.UNKNOWN;
	specRealisticSeeder.additionalTankFillType = FillType.UNKNOWN;
	specRealisticSeeder.additionalTankFillLevel = 0;
	specRealisticSeeder.additionalTank = nil;
	specRealisticSeeder.hasStoppedWorkingAnimation = false;
	specRealisticSeeder.timerDuration = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.sowingMachine#timerDuration"), 40);
	
	local seedFillUnitIndex = RealisticSeederUtil.getSeedFillUnitIndex(self.xmlFile, self, specFillUnit);
	
	if RealisticSeederUtil.foundFunctionAdditionalTank(self.xmlFile) and string.find(g_fillTypeManager:getFillTypeByIndex(specFillUnit.fillUnits[seedFillUnitIndex].fillType).name, "SEED") then
		self:setFillUnitForcedMaterialFillType(seedFillUnitIndex, FillType.SEEDS);
	end;

	if specSowingMachine ~= nil and g_fillTypeManager:getFillTypeByIndex(specFillUnit.fillUnits[specSowingMachine.fillUnitIndex].fillType).name == "SEED_POTATO" then
		g_animationManager:setFillType(specSowingMachine.animationNodes, FillType.SEEDS);
	end;
end;

function RealisticSeeder:loadSeedFillTypes(modDesc, xmlFile, storeCategory, width)
	local specSowingMachine = RealisticSeederUtil.getSpecByName(self, "sowingMachine");
	local specFillUnit = RealisticSeederUtil.getSpecByName(self, "fillUnit");

	local fillTypeNames = getXMLString(modDesc, "modDesc.realisticSeeder#fillTypes");
	local fillTypes = g_fillTypeManager:getFillTypesByNames(fillTypeNames, "Warning: '" .. self.configFileName .. "' has invalid fillType '%s'.");

	local fillUnitConfigurationId = Utils.getNoNil(self.configurations["fillUnit"], 1);
	local baseKey = string.format("vehicle.fillUnit.fillUnitConfigurations.fillUnitConfiguration(%d).fillUnits", fillUnitConfigurationId -1);

	local fillUnitNumber = 0;

    while true do
		local fillUnitKey = string.format("%s.fillUnit(%d)", baseKey, fillUnitNumber);
		
        if not hasXMLProperty(xmlFile, fillUnitKey) then
            break;
		end;

		local defaultFillType = getXMLString(xmlFile, fillUnitKey .. "#fillTypes");

		--## fix unloading width for all sowing machines! Thanks to GIANTS for this issue!

		if defaultFillType ~= nil and string.find(string.lower(defaultFillType), "seeds") then
			--## clear GIANTS shit out of table or create it, if it is missing! e.g.: potato planters!
			specFillUnit.unloading = {};

			--## set new values for table!
			unloading = {};
			unloading.node = I3DUtil.indexToObject(self.components, getXMLString(xmlFile, fillUnitKey .. "#node")) or self.rootNode;
			unloading.width = width + 18;
			unloading.offset = StringUtil.getVectorNFromString(getXMLString(xmlFile, fillUnitKey .. "#offset") or "0 0 0", 3);

			--## insert new values in table!
			table.insert(specFillUnit.unloading, unloading);
		end;
		
		if fillTypes ~= nil then
			for _, fillType in pairs(fillTypes) do
				local seedsFillType = g_fillTypeManager:getFillTypeByIndex(fillType);

				if specSowingMachine ~= nil and specSowingMachine.seeds ~= nil and defaultFillType ~= nil and string.find(string.lower(defaultFillType), "seeds") then	
					for _, fruitType in pairs(specSowingMachine.seeds) do
						local seedsFruitType = g_fillTypeManager:getFillTypeByIndex(g_fruitTypeManager:getFillTypeIndexByFruitTypeIndex(fruitType));

						if seedsFruitType ~= nil then
							--## check if the fruit type is avaiable
							if string.sub(seedsFillType.name, 6) == seedsFruitType.name and not specFillUnit.fillUnits[fillUnitNumber + 1].supportedFillTypes[fillType] then
								specFillUnit.fillUnits[fillUnitNumber + 1].supportedFillTypes[fillType] = true;

								RealisticSeederUtil.printDebug("Insert fill type " .. seedsFillType.name .. " to " .. self.configFileName, RealisticSeeder.debugPriority, true, "RealisticSeeder");
							end;
						end;
					end;
				else
					if RealisticSeederUtil.foundFunctionAdditionalTank(xmlFile) then
						local seedsFruitType = g_fruitTypeManager:getFruitTypeByName(string.sub(seedsFillType.name, 6));

						if seedsFruitType ~= nil then
							--## difference between the store categories

							if storeCategory == "seeders" or storeCategory == "planters" then
								if #specFillUnit.fillUnits == 1 then
									--## check if fertilizer and seeds is avaiable in the additional tank, if not, add it!

									for _, fillTypeToCheck in pairs({"FERTILIZER", "SEEDS"}) do
										if not specFillUnit.fillUnits[fillUnitNumber + 1].supportedFillTypes[FillType[fillTypeToCheck]] then
											specFillUnit.fillUnits[fillUnitNumber + 1].supportedFillTypes[FillType[fillTypeToCheck]] = true;

											RealisticSeederUtil.printDebug("Insert fill type " .. g_fillTypeManager:getFillTypeByIndex(FillType[fillTypeToCheck]).name .. " to " .. self.configFileName, RealisticSeeder.debugPriority, true, "RealisticSeeder");
										end;
									end;
								end;

								for _, category in pairs({"SOWINGMACHINE", "PLANTER"}) do
									local categoryIndex = g_fruitTypeManager.categories[category];
									local categoryFruitTypes = g_fruitTypeManager.categoryToFruitTypes[categoryIndex];

        							if categoryFruitTypes ~= nil then
        							    for _, fruitType in ipairs(categoryFruitTypes) do
											if string.sub(seedsFillType.name, 6) == g_fillTypeManager:getFillTypeByIndex(g_fruitTypeManager:getFillTypeIndexByFruitTypeIndex(fruitType)).name and not specFillUnit.fillUnits[fillUnitNumber + 1].supportedFillTypes[fillType] then
												specFillUnit.fillUnits[fillUnitNumber + 1].supportedFillTypes[fillType] = true;

												RealisticSeederUtil.printDebug("Insert fill type " .. seedsFillType.name .. " to " .. self.configFileName, RealisticSeeder.debugPriority, true, "RealisticSeeder");
											end;
        							    end;
									end;
								end;
							else
								--## in case, someone make an additional tank for potato... Unreal, but who knows..
								if string.sub(seedsFillType.name, 6) == "POTATO" and not specFillUnit.fillUnits[fillUnitNumber + 1].supportedFillTypes[fillType] then
									specFillUnit.fillUnits[fillUnitNumber + 1].supportedFillTypes[fillType] = true;

									RealisticSeederUtil.printDebug("Insert fill type " .. seedsFillType.name .. " to " .. self.configFileName, RealisticSeeder.debugPriority, true, "RealisticSeeder");
								end;
							end;
						end;
					end;
				end;
			end;
		end;

		fillUnitNumber = fillUnitNumber + 1;
	end;
end;

function RealisticSeeder:addFillUnitFillLevel(superFunc, farmId, fillUnitIndex, fillLevelDelta, fillType, toolType, fillInfo)
	local specSowingMachine = RealisticSeederUtil.getSpecByName(self, "sowingMachine");
	local specFillUnit = RealisticSeederUtil.getSpecByName(self, "fillUnit");

	if self:getFillUnitFillType(fillUnitIndex) ~= FillType.UNKNOWN and self:getFillUnitFillType(fillUnitIndex) ~= fillType then
		if specFillUnit.fillUnits[fillUnitIndex].fillLevel / math.max(specFillUnit.fillUnits[fillUnitIndex].capacity, 0.0001) > self:getFillTypeChangeThreshold() then
			return 0;
		end;
	end;
	
	local fillTypeDesc = g_fillTypeManager:getFillTypeByIndex(fillType);

	if fillTypeDesc ~= nil then
		if fillUnitIndex == specSowingMachine.fillUnitIndex then
			if string.sub(fillTypeDesc.name, 1, 5) == "SEED_" then
				if fillTypeDesc.name ~= "SEED_POTATO" then	
					self:setFillUnitForcedMaterialFillType(fillUnitIndex, FillType.SEEDS);
				end;
			else
				--## force material of fillVolume to be seeds, if fillType is accepted
				if self:getFillUnitSupportsFillType(fillUnitIndex, fillType) then
					self:setFillUnitForcedMaterialFillType(fillUnitIndex, FillType.SEEDS);
				end;

				--## switch from seeds to "fill type", if possible
				local fruitType = specSowingMachine.seeds[specSowingMachine.currentSeed];

				if fruitType ~= nil then
					local seedsFillType = g_fruitTypeManager:getFillTypeIndexByFruitTypeIndex(fruitType);

					if seedsFillType ~= nil then
						if self:getFillUnitSupportsFillType(fillUnitIndex, seedsFillType) then
							self:setFillUnitForcedMaterialFillType(fillUnitIndex, seedsFillType);
						end;
					end;
				end;
			end;
		end;
	end;

    return superFunc(self, farmId, fillUnitIndex, fillLevelDelta, fillType, toolType, fillInfo);
end;

SowingMachine.addFillUnitFillLevel = RealisticSeeder.addFillUnitFillLevel; --## need an hardcore overwrite... Otherwhise, the fill type will automaticlly set to "seeds"!

function RealisticSeeder:getAttachedAdditionalTanks(rootVehicle)
	local specRealisticSeeder = RealisticSeederUtil.getSpecByName(self, "realisticSeeder");
	
	if rootVehicle ~= nil then
		local specAttacherJoints = RealisticSeederUtil.getSpecByName(rootVehicle, "attacherJoints");

		if specAttacherJoints ~= nil then
			for _, implement in pairs(specAttacherJoints.attachedImplements) do
				local additionalTank = implement.object;
				local additionalTank_specAttachable = RealisticSeederUtil.getSpecByName(additionalTank, "attachable");

				if additionalTank_specAttachable.attacherVehicle ~= nil then
					local attacherJointVehicleSpec = RealisticSeederUtil.getSpecByName(additionalTank_specAttachable.attacherVehicle, "attacherJoints");
					
					if RealisticSeederUtil.foundFunctionAdditionalTank(additionalTank.xmlFile) then
						local additionalTank_specFillUnit = RealisticSeederUtil.getSpecByName(additionalTank, "fillUnit");
						local seedFillUnitIndex = RealisticSeederUtil.getSeedFillUnitIndex(additionalTank.xmlFile, additionalTank, additionalTank_specFillUnit);

						if specRealisticSeeder.additionalTankFillType == FillType.UNKNOWN then
							specRealisticSeeder.additionalTankFillType = additionalTank_specFillUnit.fillUnits[seedFillUnitIndex].fillType;
						end

						if string.find(g_fillTypeManager:getFillTypeByIndex(additionalTank_specFillUnit.fillUnits[seedFillUnitIndex].fillType).name, "SEED") then
 							specRealisticSeeder.additionalTank = additionalTank;
							specRealisticSeeder.additionalTankFillLevel = additionalTank_specFillUnit.fillUnits[seedFillUnitIndex].fillLevel;

							local seedsFillType = g_fillTypeManager:getFillTypeByName(string.sub(g_fillTypeManager:getFillTypeByIndex(additionalTank_specFillUnit.fillUnits[seedFillUnitIndex].fillType).name, 6));

							if seedsFillType ~= nil then
								additionalTank:setFillUnitFillTypeToDisplay(seedFillUnitIndex, seedsFillType.index, true)
								additionalTank:setFillUnitForcedMaterialFillType(seedFillUnitIndex, FillType.SEEDS);
							end;

							if specRealisticSeeder.additionalTankFillLevel <= 0 then
								specRealisticSeeder.additionalTankFillType = FillType.UNKNOWN;
								specRealisticSeeder.additionalTank = nil;
							end;

							break;
						else
							specRealisticSeeder.additionalTank = nil;
							specRealisticSeeder.additionalTankFillLevel = 0;

							additionalTank:setFillUnitFillTypeToDisplay(seedFillUnitIndex, g_fillTypeManager:getFillTypeByIndex(additionalTank_specFillUnit.fillUnits[seedFillUnitIndex].fillType).index, true);
							additionalTank:setFillUnitForcedMaterialFillType(seedFillUnitIndex, g_fillTypeManager:getFillTypeByIndex(additionalTank_specFillUnit.fillUnits[seedFillUnitIndex].fillType).index);
						end;
					else
						specRealisticSeeder.additionalTankFillType = FillType.UNKNOWN;
						specRealisticSeeder.additionalTank = nil;
						specRealisticSeeder.additionalTankFillLevel = 0;
					end;

					if additionalTank.getAttachedAdditionalTanks ~= nil then
						additionalTank:getAttachedAdditionalTanks(additionalTank);
					end;
				end;
			end;
		end;
	else
		specRealisticSeeder.additionalTankFillType = FillType.UNKNOWN;
		specRealisticSeeder.additionalTank = nil;
		specRealisticSeeder.additionalTankFillLevel = 0;
	end;
end;

function RealisticSeeder:onUpdateTick(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
	if not RealisticSeederUtil.getIsValidStoreCategory(getXMLString(self.xmlFile, "vehicle.storeData.category")) then
		return;
	end;
	
	local specRealisticSeeder = RealisticSeederUtil.getSpecByName(self, "realisticSeeder");
	local specSowingMachine = RealisticSeederUtil.getSpecByName(self, "sowingMachine");
	local specFillUnit = RealisticSeederUtil.getSpecByName(self, "fillUnit");
	
	self:getAttachedAdditionalTanks(self:getRootVehicle());

	if specRealisticSeeder.additionalTankFillLevel <= 0 then
		specRealisticSeeder.additionalTankFillType = FillType.UNKNOWN;
	end;

	if specSowingMachine ~= nil then
		local selectedSeedFruitType = specFillUnit.fillUnits[specSowingMachine.fillUnitIndex].fillType;

		if specRealisticSeeder.additionalTankFillType ~= FillType.UNKNOWN then
			selectedSeedFruitType = specRealisticSeeder.additionalTankFillType;
		end;

		if selectedSeedFruitType ~= nil then
			specSowingMachine.allowsSeedChanging = selectedSeedFruitType == FillType.SEEDS; --## with default seeds you can switch the seed fruit types as in the base game

			if selectedSeedFruitType ~= FillType.UNKNOWN and not specSowingMachine.allowsSeedChanging and specRealisticSeeder.oldSeedFruitType ~= selectedSeedFruitType then
				local fillTypeDesc = g_fillTypeManager:getFillTypeByIndex(selectedSeedFruitType);

				if fillTypeDesc ~= nil then
					local fruitTypeDesc = g_fruitTypeManager:getFruitTypeByName(string.sub(fillTypeDesc.name, 6));

					if fruitTypeDesc ~= nil then
						self:setSeedFruitType(fruitTypeDesc.index, true);
						
						specRealisticSeeder.oldSeedFruitType = selectedSeedFruitType;
					end;
				end;
			end;

			if self:getIsTurnedOn() then
				if self:getIsImplementChainLowered() then
					if specRealisticSeeder.hasStoppedWorkingAnimation then	
						if specRealisticSeeder.timer > 0 then
							specRealisticSeeder.timer = specRealisticSeeder.timer - 1;
						else
							g_animationManager:startAnimations(specSowingMachine.animationNodes);

							specRealisticSeeder.hasStoppedWorkingAnimation = false;
						end;
					end;
				else
					if not specRealisticSeeder.hasStoppedWorkingAnimation then
						g_animationManager:stopAnimations(specSowingMachine.animationNodes);

						specRealisticSeeder.hasStoppedWorkingAnimation = true;
						specRealisticSeeder.timer = specRealisticSeeder.timerDuration;
					end;
				end;
			end;

			if g_fillTypeManager:getFillTypeByIndex(selectedSeedFruitType).name == "SEED_POTATO" then
				g_animationManager:setFillType(specSowingMachine.animationNodes, FillType.SEEDS);
			end;
		end;
	end;
end;

function RealisticSeeder:onStartWorkAreaProcessing(superFunc, dt)
	local specRealisticSeeder = RealisticSeederUtil.getSpecByName(self, "realisticSeeder");
    local specSowingMachine = RealisticSeederUtil.getSpecByName(self, "sowingMachine");
	
	specSowingMachine.isWorking = false;
    specSowingMachine.isProcessing = false;
	
	local seedsFruitType = specSowingMachine.seeds[specSowingMachine.currentSeed];
    local dx, _, dz = localDirectionToWorld(specSowingMachine.directionNode, 0, 0, 1);
    local angleRad = MathUtil.getYRotationFromDirection(dx, dz);
    local desc = g_fruitTypeManager:getFruitTypeByIndex(seedsFruitType);
	
	if desc ~= nil and desc.directionSnapAngle ~= 0 then
        angleRad = math.floor(angleRad / desc.directionSnapAngle + 0.5) * desc.directionSnapAngle;
    end;
	
	local angle = FSDensityMapUtil.convertToDensityMapAngle(angleRad, g_currentMission.terrainDetailAngleMaxValue);
	local seedsVehicle, seedsVehicleFillUnitIndex;
	
	if specRealisticSeeder.additionalTankFillType == FillType.UNKNOWN then
		if self:getFillUnitFillLevel(specSowingMachine.fillUnitIndex) > 0 then
    	    seedsVehicle = self;
			seedsVehicleFillUnitIndex = specSowingMachine.fillUnitIndex;
		end;
	else
		if specRealisticSeeder.additionalTank ~= nil then
			local fillType = specRealisticSeeder.additionalTankFillType;

			local seedFillUnitIndex = RealisticSeederUtil.getSeedFillUnitIndex(specRealisticSeeder.additionalTank.xmlFile, specRealisticSeeder.additionalTank, specRealisticSeeder.additionalTank.spec_fillUnit);

			if specRealisticSeeder.additionalTank:getFillUnitFillLevel(seedFillUnitIndex) > 0 and specRealisticSeeder.additionalTank:getFillUnitFillType(seedFillUnitIndex) == fillType then
        		seedsVehicle = specRealisticSeeder.additionalTank;
				seedsVehicleFillUnitIndex = seedFillUnitIndex;
        	end;
		end;
	end;

    if seedsVehicle ~= nil and seedsVehicle ~= self then
		local fillType = g_fruitTypeManager:getFillTypeIndexByFruitTypeIndex(seedsFruitType);
		
        seedsVehicle:setFillUnitFillTypeToDisplay(seedsVehicleFillUnitIndex, fillType);
	end;
	
    --## check if selected fruit is available in current map
	local canFruitBePlanted = false;
	
    if g_currentMission.fruits ~= nil then
        if g_currentMission.fruits[seedsFruitType] ~= nil then
            if g_currentMission.fruits[seedsFruitType].id ~= nil and g_currentMission.fruits[seedsFruitType].id ~= 0 then
                canFruitBePlanted = true;
            end;
        end;
	end;
	
    if specSowingMachine.showWrongFruitForMissionWarning then
        specSowingMachine.showWrongFruitForMissionWarning = false;
	end;
	
    specSowingMachine.showFruitCanNotBePlantedWarning = not canFruitBePlanted;
    specSowingMachine.workAreaParameters.isActive = not specSowingMachine.needsActivation or self:getIsTurnedOn();
    specSowingMachine.workAreaParameters.canFruitBePlanted = canFruitBePlanted;
    specSowingMachine.workAreaParameters.seedsFruitType = seedsFruitType;
    specSowingMachine.workAreaParameters.angle = angle;
    specSowingMachine.workAreaParameters.seedsVehicle = seedsVehicle;
    specSowingMachine.workAreaParameters.seedsVehicleFillUnitIndex = seedsVehicleFillUnitIndex;
    specSowingMachine.workAreaParameters.lastTotalArea = 0;
    specSowingMachine.workAreaParameters.lastChangedArea = 0;
    specSowingMachine.workAreaParameters.lastStatsArea = 0;
end;

SowingMachine.onStartWorkAreaProcessing = RealisticSeeder.onStartWorkAreaProcessing; --## need an hardcore overwrite... Otherwhise, the seed fill type will automaticlly set to "seeds"!

function RealisticSeeder:unloadFillUnits(superFunc, ignoreWarning)
	if not self.isServer then
		g_client:getServerConnection():sendEvent(FillUnitUnloadEvent:new(self));
	else
		local specFillUnit = RealisticSeederUtil.getSpecByName(self, "fillUnit");
		local unloadingPlaces = specFillUnit.unloading;
		local places = {};

		for _, unloading in ipairs(unloadingPlaces) do
			local node = unloading.node;
			local ox, oy, oz = unpack(unloading.offset);
			local x, y, z = localToWorld(node, ox - unloading.width * 0.1, oy, oz);
			local place = {
				startZ = z,
				startY = y,
				startX = x
			};

			place.rotX, place.rotY, place.rotZ = getWorldRotation(node);
			place.dirX, place.dirY, place.dirZ = localDirectionToWorld(node, 1, 0, 0);
			place.dirPerpX, place.dirPerpY, place.dirPerpZ = localDirectionToWorld(node, 0, 0, 1);
			place.yOffset = 1;
			place.maxWidth = math.huge;
			place.maxLength = math.huge;
			place.width = unloading.width;

			table.insert(places, place);
		end;

		local usedPlaces = {};
		local success = true;
		local availablePallets = {};

		for fillUnitNumber, fillUnit in ipairs(self:getFillUnits()) do
			local fillLevel = self:getFillUnitFillLevel(fillUnitNumber);
			local fillTypeIndex = self:getFillUnitFillType(fillUnitNumber);
			local fillType = g_fillTypeManager:getFillTypeByIndex(fillTypeIndex);

			if fillUnit.canBeUnloaded and fillLevel > 0 and fillType.palletFilename ~= nil then
				while fillLevel > 0 do
					local pallet = availablePallets[fillTypeIndex];
					local isNewPallet = false;

					if pallet == nil then
						local sizeWidth, sizeLength, widthOffset, lengthOffset = StoreItemUtil.getSizeValues(fillType.palletFilename, "vehicle", 0, {});
						local x, _, z, place, width, _ = PlacementUtil.getPlace(places, sizeWidth, sizeLength, widthOffset, lengthOffset, usedPlaces, true, true, true);

						if x == nil then
							success = false;

							break;
						end;

						PlacementUtil.markPlaceUsed(usedPlaces, place, width);

						pallet = g_currentMission:loadVehicle(fillType.palletFilename, x, nil, z, 0, place.rotY, true, 0, Vehicle.PROPERTY_STATE_OWNED, self:getOwnerFarmId(), nil, nil);

						pallet:emptyAllFillUnits(true);

						isNewPallet = true;
					end

					local fillUnitIndex = pallet:getFirstValidFillUnitToFill(fillTypeIndex);

					if fillUnitIndex ~= nil then
						local appliedDelta = pallet:addFillUnitFillLevel(self:getOwnerFarmId(), fillUnitIndex, fillLevel, fillTypeIndex, ToolType.UNDEFINED, nil);

						self:addFillUnitFillLevel(self:getOwnerFarmId(), fillUnitNumber, -appliedDelta, fillTypeIndex, ToolType.UNDEFINED, nil);

						fillLevel = fillLevel - appliedDelta;
					else
						if isNewPallet then
							break;
						end;

						availablePallets[fillTypeIndex] = nil;
					end;
				end;
			end;
		end;

		if (ignoreWarning == nil or not ignoreWarning) and not success then
			g_currentMission:addIngameNotification(FSBaseMission.INGAME_NOTIFICATION_INFO, g_i18n:getText("fillUnit_unload_nospace"));
		end;

		return success;
	end;
end;

FillUnit.unloadFillUnits = RealisticSeeder.unloadFillUnits;

function RealisticSeeder:updateFillLevelFrames(superFunc)
	local _, yOffset = self:getPosition();

	for fillLevelBufferNumber = 1, #self.fillLevelBuffer do
		local fillLevelInformation = self.fillLevelBuffer[fillLevelBufferNumber];

		if self.fillTypeFrames[fillLevelInformation.fillType] ~= nil then 
			local value = 0;

			if fillLevelInformation.capacity > 0 then
				value = fillLevelInformation.fillLevel / fillLevelInformation.capacity;
			end;

			local frame = self.fillTypeFrames[fillLevelInformation.fillType];

			frame:setVisible(true);

			local fillBar = self.fillTypeLevelBars[fillLevelInformation.fillType];
			local _, yScale = fillBar:getScale();

			fillBar:setScale(value * self.uiScale, yScale);

			local posX, posY = frame:getPosition();

			frame:setPosition(posX, yOffset);

			local fillText = string.format("%d (%d%%)", MathUtil.round(fillLevelInformation.fillLevel), 100 * value);

			self.fillLevelTextBuffer[#self.fillLevelTextBuffer + 1] = fillText;

			if fillLevelInformation.fillType ~= FillType.UNKNOWN then
				local fillTypeText = g_fillTypeManager:getFillTypeByIndex(fillLevelInformation.fillType).title;
				self.fillTypeTextBuffer[#self.fillLevelTextBuffer] = fillTypeText;
			end

			yOffset = yOffset + self.frameHeight;
		else
			--## fix, create fill type frame when nil, instead of spamming the log file full..
			local posX, posY = self:getPosition();
			local frame = self:createFillTypeFrame(self.hudAtlasPath, posX, posY, g_fillTypeManager:getFillTypeByIndex(fillLevelInformation.fillType));

			self.fillTypeFrames[g_fillTypeManager:getFillTypeByIndex(fillLevelInformation.fillType).index] = frame;

			frame:setScale(self.uiScale, self.uiScale);

			self:addChild(frame);
		end;
	end;
end;

FillLevelsDisplay.updateFillLevelFrames = Utils.overwrittenFunction(FillLevelsDisplay.updateFillLevelFrames, RealisticSeeder.updateFillLevelFrames);

function RealisticSeeder:updateFillLevelBuffers(superFunc)
	local function clearTable(table)
		for tableNumber = #table, 1, -1 do
			table[tableNumber] = nil;
		end;
	end;
	
	clearTable(self.fillLevelTextBuffer);
    clearTable(self.fillLevelBuffer);
	clearTable(self.fillTypeTextBuffer);
	
    for _, frame in pairs(self.fillTypeFrames) do
		if frame ~= nil and frame.setVisible ~= nil then	
			frame:setVisible(false);
		end;
	end;
	
    self.vehicle:getFillLevelInformation(self.fillLevelBuffer);
end;

FillLevelsDisplay.updateFillLevelBuffers = Utils.overwrittenFunction(FillLevelsDisplay.updateFillLevelBuffers, RealisticSeeder.updateFillLevelBuffers);

--## Hook loadSpecValueFillTypes function from FillUnit.lua to add potatos, fertilizer and seeds, if it necessary, in the shop view
FillUnit.loadSpecValueFillTypes = function(xmlFile, customEnvironment)
	--print("xmlFile = " .. tostring(xmlFile) .. " customEnvironment = " .. tostring(customEnvironment));
	
	local fillTypeNames, fillTypeCategoryNames, fillTypes, fruitTypeNames = nil;
	local rootName = getXMLRootName(xmlFile);
	local fillUnitConfigurationNumber = 0;
	local foundHolmerPotatoReadyMod = false;

	for _, mod in pairs(g_modManager.mods) do
		if mod.title == "Holmer Potato Ready" then	
			if g_modIsLoaded[tostring(mod.modName)] then
				foundHolmerPotatoReadyMod = true;

				break;
			end;
		end;
	end;

	while true do
		local fillUnitConfigurationKey = string.format(rootName .. ".fillUnit.fillUnitConfigurations.fillUnitConfiguration(%d)", fillUnitConfigurationNumber);

		if not hasXMLProperty(xmlFile, fillUnitConfigurationKey) then
			break;
		end;

		local fillUnitNumber = 0;

		while true do
			local fillUnitKey = string.format(fillUnitConfigurationKey .. ".fillUnits.fillUnit(%d)", fillUnitNumber);

			if not hasXMLProperty(xmlFile, fillUnitKey) then
				break;
			end;

			local showInShop = getXMLBool(xmlFile, fillUnitKey .. "#showInShop");
			local capacity = getXMLFloat(xmlFile, fillUnitKey .. "#capacity");

			if (showInShop == nil or showInShop) and (capacity == nil or capacity > 0) then
				local currentFillTypes = getXMLString(xmlFile, fillUnitKey .. "#fillTypes");

				if currentFillTypes ~= nil then
					if string.find(currentFillTypes, "seeds") and not string.find(currentFillTypes, "fertilizer") then
						currentFillTypes = currentFillTypes .. " fertilizer"; --## add fertilizer to the shop view
					elseif string.find(currentFillTypes, "fertilizer") and not string.find(currentFillTypes, "seeds") then
						currentFillTypes = currentFillTypes .. "seeds"; --## add seeds to the shop view
					end;

					if foundHolmerPotatoReadyMod then
						if string.find(currentFillTypes, "sugarBeet") and not string.find(currentFillTypes, "potato") then
							currentFillTypes = currentFillTypes .. " potato"; --## add potato to the shop view
						end;
					end;

					if fillTypeNames == nil then
						fillTypeNames = currentFillTypes;
					else
						fillTypeNames = fillTypeNames .. " " .. currentFillTypes;
					end;
				end;

				local currentFillTypeCategories = getXMLString(xmlFile, fillUnitKey .. "#fillTypeCategories");
				
				if RealisticSeederUtil.foundFunctionAdditionalTank(xmlFile) then
					currentFillTypeCategories = nil;

					fillTypeNames = "fertilizer seeds";
				end;

				if currentFillTypeCategories ~= nil then
					if fillTypeCategoryNames == nil then
						fillTypeCategoryNames = currentFillTypeCategories;
					else
						fillTypeCategoryNames = fillTypeCategoryNames .. " " .. currentFillTypeCategories;
					end;
				end;
			end;

			fillUnitNumber = fillUnitNumber + 1;
		end;

		fillUnitConfigurationNumber = fillUnitConfigurationNumber + 1;
	end;

	if fillTypeNames == nil then
		fillTypeNames = getXMLString(xmlFile, rootName .. ".fillTypes");
	end;

	fillTypeNames = Utils.getNoNil(getXMLString(xmlFile, rootName .. ".storeData.specs.fillTypes"), fillTypeNames);

	if fillTypeNames ~= nil then
		fillTypeCategoryNames = nil;
	end;

	if fillTypeCategoryNames == nil then
		fillTypeCategoryNames = getXMLString(xmlFile, rootName .. ".fillTypeCategories");
	end;

	if fillTypes == nil then
		fruitTypeNames = getXMLString(xmlFile, rootName .. ".fruitTypes");
	end;

	if fillTypes == nil then
		fruitTypeNames = getXMLString(xmlFile, rootName .. ".cutter#fruitTypes");

		if fruitTypeNames ~= nil and foundHolmerPotatoReadyMod then
			if string.find(fruitTypeNames, "sugarBeet") and not string.find(fruitTypeNames, "potato") then
				fruitTypeNames = fruitTypeNames .. " potato"; --## add potato to the shop view
			end;
		end;
	end;

	local fruitTypeCategoryNames = getXMLString(xmlFile, rootName .. ".cutter#fruitTypeCategories");
	local useWindrowed = Utils.getNoNil(getXMLBool(xmlFile, rootName .. ".cutter#useWindrowed"), false);

	fillTypeCategoryNames = Utils.getNoNil(getXMLString(xmlFile, rootName .. ".storeData.specs.fillTypeCategories"), fillTypeCategoryNames);

	return {
		categoryNames = fillTypeCategoryNames,
		fillTypeNames = fillTypeNames,
		fruitTypeNames = fruitTypeNames,
		fruitTypeCategoryNames = fruitTypeCategoryNames,
		useWindrowed = useWindrowed
	};
end;